from collections.abc import Iterable

import markov_clustering as mc
import networkx as nx

from matchescu.similarity import SimilarityGraph

from matchescu.clustering._base import T, ClusteringAlgorithm


class MarkovClustering(ClusteringAlgorithm[T]):
    def __init__(
        self,
        all_refs: Iterable[T],
        threshold: float = 0.0,
        inflation_power: float = 2.0,
        expansion_power: int = 2,
        prune_threshold: float = 0.001,
    ):
        super().__init__(all_refs, threshold)
        self._item_to_index = {item: idx for idx, item in enumerate(self._items)}
        self._expansion_power = expansion_power
        self._inflation_power = inflation_power
        self._prune_threshold = prune_threshold

    def __call__(self, similarity_graph: SimilarityGraph) -> frozenset[frozenset[T]]:
        g = nx.DiGraph()

        for u, v in similarity_graph.matches():
            g.add_edge(u, v, weight=similarity_graph.weight(u, v))

        linked_ref_ids = list(g.nodes)
        adj_matrix = nx.to_numpy_array(g, weight="weight")
        adj_matrix[adj_matrix < 0] = 0

        result = mc.run_mcl(
            adj_matrix,
            self._expansion_power,
            self._inflation_power,
            pruning_threshold=self._prune_threshold,
        )
        cluster_indices = mc.get_clusters(result)

        singletons = set(self._items) - set(linked_ref_ids)
        clusters = set()
        for c in cluster_indices:
            cluster = frozenset(linked_ref_ids[idx] for idx in c)
            singletons.difference_update(cluster)
            clusters.add(cluster)
        for singleton in singletons:
            clusters.add(frozenset([singleton]))

        return frozenset(clusters)



import itertools
from typing import Iterable

import networkx as nx
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import pdist

from matchescu.similarity import SimilarityGraph

from matchescu.clustering._base import ClusteringAlgorithm, T


class HierarchicalAgglomerativeClustering(ClusteringAlgorithm[T]):
    def __init__(
        self,
        all_refs: Iterable[T],
        distance_function: str = "cosine",
        max_cluster_distance: float = 1.0,
    ) -> None:
        super().__init__(all_refs, 0.0)
        self._fcluster_threshold = max_cluster_distance
        self._distance_function = distance_function
        self._linkage_method = "ward"
        self._clustering_criterion = "distance"

    def _distance_matrix(self, similarity_graph: SimilarityGraph) -> np.ndarray:
        g = nx.DiGraph()
        g.add_nodes_from(self._items)
        g.add_weighted_edges_from(
            itertools.starmap(
                lambda u, v, data: (u, v, data.get("weight", 0.0)),
                similarity_graph.edges,
            )
        )
        sim_matrix = nx.to_numpy_array(
            g, nodelist=self._items, weight="weight"
        ) + np.eye(len(self._items))
        return 1 - ((sim_matrix + sim_matrix.T) / 2)

    def __call__(self, similarity_graph: SimilarityGraph) -> frozenset[frozenset[T]]:
        distance_matrix = self._distance_matrix(similarity_graph)

        # compute hierarchical clusters based on average
        condensed_distance_matrix = pdist(distance_matrix, self._distance_function)
        Z = linkage(condensed_distance_matrix, method=self._linkage_method)

        # flatten the clusters based on distance
        cluster_assignments = fcluster(
            Z, self._fcluster_threshold, criterion=self._clustering_criterion
        )

        # map cluster assignments back to items
        unique_clusters = np.unique(cluster_assignments)
        return frozenset(
            frozenset(self._items[idx] for idx in np.where(cluster_assignments == c)[0])
            for c in unique_clusters
        )



import abc
from collections.abc import Iterable
from typing import TypeVar, Hashable, Generic

from matchescu.similarity import SimilarityGraph

T = TypeVar("T", bound=Hashable)


class ClusteringAlgorithm(Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, all_refs: Iterable[T], threshold: float) -> None:
        self._items = list(set(all_refs))
        self._threshold = threshold

    @abc.abstractmethod
    def __call__(self, similarity_graph: SimilarityGraph) -> frozenset[frozenset[T]]:
        pass



from matchescu.clustering._corr import WeightedCorrelationClustering
from matchescu.clustering._wcc import WeaklyConnectedComponents
from matchescu.clustering._ecp import EquivalenceClassPartitioner
from matchescu.clustering._mcl import MarkovClustering


__all__ = [
    "EquivalenceClassPartitioner",
    "MarkovClustering",
    "WeaklyConnectedComponents",
    "WeightedCorrelationClustering",
]



from collections import defaultdict
from collections.abc import Iterable

import networkx as nx
from matchescu.similarity import SimilarityGraph

from matchescu.clustering._base import T, ClusteringAlgorithm


class ParentCenterClustering(ClusteringAlgorithm[T]):
    def __init__(self, all_refs: Iterable[T], threshold: float = 0.0) -> None:
        super().__init__(all_refs, threshold)

    @staticmethod
    def _find_root(parents, node):
        """
        Finds the root parent of a node in the parent dictionary.
        """
        if parents[node] == node:
            return node

        return ParentCenterClustering._find_root(parents, parents[node])

    def __call__(self, similarity_graph: SimilarityGraph) -> frozenset[frozenset[T]]:
        graph = nx.DiGraph()
        graph.add_nodes_from(self._items)
        graph.add_edges_from(similarity_graph.matches())

        parents = {node: node for node in self._items}
        updated = True

        while updated:
            updated = False
            new_parents = parents.copy()
            for node in self._items:
                best_parent = node
                max_similarity = -1

                for predecessor in graph.predecessors(node):
                    similarity = similarity_graph.weight(predecessor, node)
                    if similarity > max_similarity:
                        max_similarity = similarity
                        best_parent = predecessor

                # Update parent to the parent of the best predecessor
                if best_parent != node:
                    root_parent = self._find_root(parents, best_parent)
                    if new_parents[node] != root_parent:
                        new_parents[node] = root_parent
                        updated = True

            parents = new_parents

        result = defaultdict(list)
        for u, v in parents.items():
            result[v].append(u)

        return frozenset(
            frozenset(node for node in cluster) for cluster in result.values()
        )
